package de.hft.stuttgart.citydoctor2.healer.genetic.gui;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.datastructure.BoundingBox;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GmlId;
import de.hft.stuttgart.citydoctor2.gui.MainWindow;
import de.hft.stuttgart.citydoctor2.gui.TriangulatedGeometry;
import de.hft.stuttgart.citydoctor2.gui.View;
import de.hft.stuttgart.citydoctor2.gui.ViewRegistration;
import de.hft.stuttgart.citydoctor2.healer.HealingMethods;
import de.hft.stuttgart.citydoctor2.healer.HealingPlan;
import de.hft.stuttgart.citydoctor2.healer.genetic.GeneticAlgorithm;
import de.hft.stuttgart.citydoctor2.healer.genetic.NoMutation;
import de.hft.stuttgart.citydoctor2.healer.genetic.OrderCrossover;
import de.hft.stuttgart.citydoctor2.healer.genetic.RandomHeuristic;
import de.hft.stuttgart.citydoctor2.healer.gui.HealerGUI;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.Copy;
import javafx.beans.property.DoubleProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.AmbientLight;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.PerspectiveCamera;
import javafx.scene.SceneAntialiasing;
import javafx.scene.SnapshotParameters;
import javafx.scene.SubScene;
import javafx.scene.control.Button;
import javafx.scene.image.Image;
import javafx.scene.input.MouseButton;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.shape.CullFace;
import javafx.scene.shape.DrawMode;
import javafx.scene.transform.Rotate;

public class GeneticsGUI extends View {
	
	private static final double CAMERA_INITIAL_X_ANGLE = 20.0;
	private static final double CAMERA_INITIAL_Y_ANGLE = 120.0;

	private Parent pane;
	private Image image;

	private DoubleProperty translateZ;
	private DoubleProperty cameraXRot;
	private DoubleProperty cameraYRot;
	private boolean enableGuiControls = true;
	
	private Group currentWorld;
	private Group nextWorld;
	private Group currentMeshGroup;
	private Group nextMeshGroup;

	@FXML
	private Pane view1;

	@FXML
	private Pane view2;

	@FXML
	private Button test;

	private Geometry currentGeometry;
	private CityObject currentFeature;
	private Checker currentChecker;
	private TriangulatedGeometry currentTriangulatedGeometry;
	private TriangulatedGeometry nextTriangulatedGeometry;
	private Geometry nextGeometry;
	private double dragX;
	private double dragY;

	public GeneticsGUI() {
		image = new Image(getClass().getResourceAsStream("genetic.png"));
	}

	public void initialize() {
		view2.prefWidthProperty().bind(view1.widthProperty());
		view2.prefHeightProperty().bind(view1.heightProperty());

		translateZ = new SimpleDoubleProperty(0);
		cameraXRot = new SimpleDoubleProperty(CAMERA_INITIAL_X_ANGLE);
		cameraYRot = new SimpleDoubleProperty(CAMERA_INITIAL_Y_ANGLE);

		test.setOnAction(ae -> {
			test.setDisable(true);
			enableGuiControls = false;
			GmlId gmlId = currentGeometry.getGmlId();
			OrderCrossover cross = new OrderCrossover();
			NoMutation mutation = new NoMutation();
			RandomHeuristic heuristic = new RandomHeuristic(HealingMethods.getAvailableHealingMethods());
			VisualFitness fitness = new VisualFitness(this, currentGeometry, currentChecker);
			GeneticAlgorithm<HealingPlan> ga = GeneticAlgorithm.of(cross, mutation, heuristic, fitness);
			ga.run();
			
			CityObject nextFeature = Copy.copy(currentFeature);
			nextGeometry = findNextGeometry(gmlId, nextFeature);
			System.out.println(gmlId);
			System.out.println(nextGeometry);
			System.out.println(nextFeature.getGeometries().get(0));
			showCurrentGeometry();
			showNextGeometry();
			test.setDisable(false);
			enableGuiControls = true;
		});

		buildCurrent3dView();
		buildNext3dView();
	}
	
	private Geometry findNextGeometry(GmlId gmlId, CityObject nextFeature) {
		Geometry[] foundGeom = new Geometry[1];
		Check geomFinder = new Check() {

			@Override
			public void check(Geometry geom) {
				if (geom.getGmlId().equals(gmlId)) {
					foundGeom[0] = geom;
				}
			}

			@Override
			public RequirementType getType() {
				return null;
			}

			@Override
			public Check createNewInstance() {
				return null;
			}

			@Override
			public CheckId getCheckId() {
				return null;
			}

			@Override
			public Set<Requirement> appliesToRequirements() {
				return null;
			}
		};
		nextFeature.accept(geomFinder);
		return foundGeom[0];
	}
	
	private void showCurrentGeometry() {
		currentTriangulatedGeometry = TriangulatedGeometry.of(currentGeometry);
		currentTriangulatedGeometry.setCullFace(CullFace.BACK);
		currentTriangulatedGeometry.setDrawMode(DrawMode.FILL);
		zoomOutForBoundingBox(currentGeometry.calculateBoundingBox());
		getCurrentMeshGroup().getChildren().clear();
		getCurrentMeshGroup().getChildren().addAll(currentTriangulatedGeometry.getMeshes());
	}
	
	private void showNextGeometry() {
		nextTriangulatedGeometry = TriangulatedGeometry.of(nextGeometry);
		nextTriangulatedGeometry.setCullFace(CullFace.BACK);
		nextTriangulatedGeometry.setDrawMode(DrawMode.FILL);
		zoomOutForBoundingBox(nextGeometry.calculateBoundingBox());
		getNextMeshGroup().getChildren().clear();
		getNextMeshGroup().getChildren().addAll(nextTriangulatedGeometry.getMeshes());

	}
	
	private Group getNextMeshGroup() {
		return nextMeshGroup;
	}

	private List<Image> makeScreenshots(int xRotSteps, int yRotSteps) {
		List<Image> images = new ArrayList<>();
		SnapshotParameters params = new SnapshotParameters();
		for (int j = 0; j < xRotSteps; j++) {
			cameraXRot.set(j * (180d / xRotSteps));
			for (int i = 0; i < yRotSteps; i++) {
				cameraYRot.set(i * (360d / yRotSteps));
				Image screen = view1.snapshot(params, null);
				images.add(screen);
			}
		}
		cameraXRot.set(0);
		cameraYRot.set(0);
		return images;
	}

	private void buildCurrent3dView() {
		Group currentRoot = new Group();
		SubScene currentScene = new SubScene(currentRoot, 500, 300, true, SceneAntialiasing.BALANCED);
		currentScene.heightProperty().bind(view1.heightProperty());
		currentScene.widthProperty().bind(view1.widthProperty());
		currentScene.setFill(Color.AZURE);
		view1.getChildren().add(currentScene);
		currentWorld = new Group();
		currentRoot.getChildren().add(currentWorld);
		currentMeshGroup = new Group();
		currentWorld.getChildren().add(currentMeshGroup);

		AmbientLight al = new AmbientLight(Color.WHITE);
		currentRoot.getChildren().add(al);

		PerspectiveCamera currentCamera = new PerspectiveCamera(true);
		currentCamera.setNearClip(0.1);
		currentCamera.setFarClip(10000d);
		currentCamera.translateZProperty().bind(translateZ);
		Rotate cameraZRotation = new Rotate();
		Rotate cameraXRotation = new Rotate();
		cameraXRotation.setAxis(Rotate.X_AXIS);
		cameraZRotation.setAxis(Rotate.Z_AXIS);
		cameraZRotation.angleProperty().bind(cameraXRot);
		cameraXRotation.angleProperty().bind(cameraYRot);
		currentWorld.getTransforms().add(cameraXRotation);
		currentWorld.getTransforms().add(cameraZRotation);
		currentRoot.getChildren().add(currentCamera);
		currentScene.setCamera(currentCamera);
		setupMeshViewControls(view1);
	}
	
	private void setupMeshViewControls(Pane canvas) {
		canvas.setOnMousePressed(me -> {
			if (enableGuiControls && me.getButton() == MouseButton.PRIMARY) {
				dragX = me.getScreenX();
				dragY = me.getScreenY();
			}
		});

		canvas.setOnScroll(se -> translateZ.set(translateZ.get() + se.getDeltaY() / 10d));

		canvas.setOnMouseDragged(me -> {
			if (enableGuiControls && me.getButton() == MouseButton.PRIMARY) {
				double deltaX = me.getScreenX() - dragX;
				double deltaY = me.getScreenY() - dragY;
				dragX = me.getScreenX();
				dragY = me.getScreenY();
				cameraXRot.set(cameraXRot.get() + ((deltaX / 3d) % 360));
				cameraYRot.set(cameraYRot.get() + ((deltaY / 3d) % 360));
			}
		});
	}

	private void buildNext3dView() {
		Group nextRoot = new Group();
		SubScene nextScene = new SubScene(nextRoot, 500, 300, true, SceneAntialiasing.BALANCED);
		nextScene.heightProperty().bind(view2.heightProperty());
		nextScene.widthProperty().bind(view2.widthProperty());
		nextScene.setFill(Color.AZURE);
		view2.getChildren().add(nextScene);
		nextWorld = new Group();
		nextRoot.getChildren().add(nextWorld);
		nextMeshGroup = new Group();
		nextWorld.getChildren().add(nextMeshGroup);

		AmbientLight al = new AmbientLight(Color.WHITE);
		nextRoot.getChildren().add(al);

		PerspectiveCamera nextCamera = new PerspectiveCamera(true);
		nextCamera.setNearClip(0.1);
		nextCamera.setFarClip(10000d);
		nextCamera.translateZProperty().bind(translateZ);
		Rotate cameraZRotation = new Rotate();
		Rotate cameraXRotation = new Rotate();
		cameraXRotation.setAxis(Rotate.X_AXIS);
		cameraZRotation.setAxis(Rotate.Z_AXIS);
		cameraZRotation.angleProperty().bind(cameraXRot);
		cameraXRotation.angleProperty().bind(cameraYRot);
		nextWorld.getTransforms().add(cameraXRotation);
		nextWorld.getTransforms().add(cameraZRotation);
		nextRoot.getChildren().add(nextCamera);
		nextScene.setCamera(nextCamera);
	}

	@Override
	public Optional<HBox> getToolbar() {
		return Optional.empty();
	}

	@Override
	public Node getMainScreen() {
		return pane;
	}

	@Override
	public Image getViewLogo() {
		return image;
	}

	@Override
	public void initializeView(MainWindow mainWindow) {
		FXMLLoader loader = new FXMLLoader(getClass().getResource("geneticGUI.fxml"));
		loader.setController(this);
		try {
			pane = loader.load();
		} catch (IOException e) {
			throw new UncheckedIOException(e);
		}
		try {
			setupViews();
		} catch (CityGmlParseException | InvalidGmlFileException e) {
			e.printStackTrace();
		}
	}

	private void setupViews() throws CityGmlParseException, InvalidGmlFileException {
		ParserConfiguration parserConfig = new ParserConfiguration(8, false);
//		CityDoctorModel model1 = CityGmlParser.parseCityGmlFile("src/test/resources/SimpleSolid_SrefBS.gml",
//				parserConfig);
		CityDoctorModel model1 = CityGmlParser.parseCityGmlFile("src/test/resources/TestBuilding3.gml",
				parserConfig);
		currentChecker = new Checker(model1);
		currentChecker.runChecks();
		Building building = model1.getBuildings().get(0);
		currentGeometry = building.getGeometries().get(0);
		currentFeature = building;
		showCurrentGeometry();
	}

	public void zoomOutForBoundingBox(BoundingBox b) {
		double cameraDistance = 2.0d;
		Vector3d objectSizes = b.getDiagonal();
		double objectSize = max(objectSizes.getX(), objectSizes.getY(), objectSizes.getZ());
		double cameraView = 2.0d * Math.tan(0.5d * Math.toRadians(30)); // Visible height 1 meter in front
		double distance = cameraDistance * objectSize / cameraView; // Combined wanted distance from the object
		distance += 0.5d * objectSize; // Estimated offset from the center to the outside of the object
		translateZ.set(-distance);
	}

	private double max(double a, double b, double c) {
		double temp = Math.max(a, b);
		return Math.max(temp, c);
	}

	public Group getCurrentMeshGroup() {
		return currentMeshGroup;
	}

	@Override
	public void onHide() {
		// TODO Auto-generated method stub

	}

	@Override
	public void onShow(CityDoctorModel model, Checker checker) {
		// TODO Auto-generated method stub

	}

	public static void main(String[] args) {
		ViewRegistration.registerView(new GeneticsGUI());
		HealerGUI.main(args);
	}

	public List<Image> createScreenshots(Geometry geom, double zoomLevel, int xRotSteps, int yRotSteps) {
		currentGeometry = geom;
		currentTriangulatedGeometry = TriangulatedGeometry.of(currentGeometry);
		currentTriangulatedGeometry.setCullFace(CullFace.BACK);
		currentTriangulatedGeometry.setDrawMode(DrawMode.FILL);
		setZoomLevel(zoomLevel);
		getCurrentMeshGroup().getChildren().clear();
		getCurrentMeshGroup().getChildren().addAll(currentTriangulatedGeometry.getMeshes());
		return makeScreenshots(xRotSteps, yRotSteps);
	}

	private void setZoomLevel(double zoomLevel) {
		translateZ.set(zoomLevel);
	}

}
